/*
  jigdo-port: a portable subset of jigdo-file      version: see VERSION below

  See below for a few #defines that may have to be adopted to your situation.

  NOTE: The only incompatibility with jigdo-file that produces NO error,
  is that jigdo-port can ONLY SCAN FILES, not directories. If you want
  recursive scanning, use "find . -type f | jigdo-port --files-from=-"

  See the jigdo-file documentation for usage info.

  Copyright (C) 2002  J.A. Bezemer  <J.A.Bezemer@opensourcepartners.nl>
  Licensed under the GNU General Public License, version 2 exactly.
  I didn't take a single look at the official jigdo-file source except
  for the base64-but-why-`-'-and-`_' details.
  If you want an example of beautiful object oriented coding, see jigdo-file.
  If you want the harsh reality of fully-portable plain-C coding, look here.
  This code was tested on a multitude of (mainly ancient) systems. Be sure
  to re-test if you make any change, however small it may seem.
*/

/* System headers. May not be in the same place everywhere */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/types.h>
/* Maybe you also need:
   #include <memory.h>
   #include <sys/file.h>
   #include <sys/fcntl.h>
*/

#include "zlib-1.1.3/zlib.h"

#define VERSION "1.0"

#define ULINT  unsigned long	/* preferably >= 6 bytes, but if image
				   is < 2GB then signed long is fine */
/* unsigned long long   is preferred on recent gcc compilers */

#define UWORD32  unsigned long	/* This probably MUST be an exactly-32-bit
				   unsigned type, needed only by md5.c */

#define HAVE_FTRUNCATE		/* Undefine this for non-POSIX systems that
				   don't have ftruncate */

#define noWORDS_BIGENDIAN	/* Define if 4-byte words are big-endian on
				   your system. */

#define md5byte  unsigned char


/* As little object & header mess as possible please. Include .c directly.

   Note: this uses memset and memcpy which may not be extremely portable,
   or may not be available in string.h. If needed, you can write your
   own simple replacements (and also for str* functions)
*/
#include "md5.c"


#define CMD_NONE	 0
#define CMD_MAKEIMAGE	 1
#define CMD_PRINTMISSING 2
#define CMD_VERIFY	 3


typedef struct _descent descent;
struct _descent
{
  descent *next;
  unsigned char type;
  ULINT len;
  ULINT pos;
  md5byte md5[16];
};

typedef struct _uricouple uricouple;
struct _uricouple
{
  uricouple *next;
  char *label;
  char *value;
};


char *imagetmp = NULL;		/* mallocd */
char *template = NULL;
int tmpimageavail = 0;
descent *deschead = NULL;
ULINT imagelen = 0;
md5byte imagemd5[16];
FILE *tmpimagefile = NULL;
int tmpimageopen = 0;
uricouple *urihead = NULL;


void
errorexit (s1, s2)
     char *s1;
     char *s2;
{
  fprintf (stderr, s1, s2);
  exit (3);
}


ULINT readint6 (f)
     FILE *
       f;
{
  ULINT res = 0;
  unsigned char c[6];

  fread (c, 1, 6, f);
  res = c[5];
  res = (res << 8) + c[4];
  res = (res << 8) + c[3];
  res = (res << 8) + c[2];
  res = (res << 8) + c[1];
  res = (res << 8) + c[0];
  return res;
}


size_t
writeint6 (f, i)
     FILE *f;
ULINT i;
/* returns 6 if written correctly */
{
  unsigned char c[6];
  ULINT res = i;

  c[0] = res % 256;
  res = res >> 8;
  c[1] = res % 256;
  res = res >> 8;
  c[2] = res % 256;
  res = res >> 8;
  c[3] = res % 256;
  res = res >> 8;
  c[4] = res % 256;
  res = res >> 8;
  c[5] = res % 256;

  return fwrite (c, 1, 6, f);
}


void
writedesc (f)
     FILE *f;
{
  descent *desccurr = deschead;
  ULINT descpos = 16;
  unsigned char one = 1;
  char *desc = "DESC";

  /* seek to DESC but skip header (write that last) */
  if (fseek (f, imagelen + 10, SEEK_SET))
    errorexit
      ("\njigdo-port: couldn't seek to DESC in temporary image `%s' (1).\n\n",
       imagetmp);

  while (desccurr != NULL)
    {
      /* No error checking here, do it later */
      descpos += 23;
      fwrite (&(desccurr->type), 1, 1, f);

      writeint6 (f, desccurr->len);
      if (desccurr->type > 2)
	fwrite (desccurr->md5, 1, 16, f);
      else
	descpos -= 16;

      desccurr = desccurr->next;
    }

  descpos += 23;
  fwrite (&one, 1, 1, f);
  writeint6 (f, imagelen);
  fwrite (imagemd5, 1, 16, f);

  /* last write: do error checking */
  if (writeint6 (f, descpos) != 6)
    errorexit
      ("\njigdo-port: couldn't write last desclen in temporary image `%s'.\n\n",
       imagetmp);

  /* seek to DESC and write header */
  if (fseek (f, imagelen, SEEK_SET))
    errorexit
      ("\njigdo-port: couldn't seek to DESC in temporary image `%s' (2).\n\n",
       imagetmp);

  fwrite (desc, 1, 4, f);
  if (writeint6 (f, descpos) != 6)
    errorexit
      ("\njigdo-port: couldn't write first desclen in temporary image `%s'.\n\n",
       imagetmp);

  fflush (f);
}


void
mkfirsttmpimage ()
{
#define BUFSIZE 30000
  unsigned char bufcomp[BUFSIZE];
  unsigned char bufuncomp[BUFSIZE];
  unsigned char *bufcompstart = NULL;
  unsigned char *bufuncompstart = NULL;
  int bufcompavail = 0;
  int bufuncompoccupied = 0;
  int i;
  ULINT written = 0;
  ULINT nextwritten = 0;
  ULINT togo;
  ULINT part;
  FILE *templatefile = NULL;
  z_stream myz;
  descent *desccurr = NULL;
  char c1 = 0, c2 = 0;
  int linesread = 0;
  ULINT bytesstillindatapart = 0;	/* compressed */
  ULINT bytesstillfromdatapart = 0;	/* uncompressed, as check */
  ULINT byteswrittentothistype2;
  int err;


  if (tmpimageavail)		/* Already done */
    return;

  /* open/check template now: no need to write temp image if no template */
  templatefile = fopen (template, "rb");
  if (!templatefile)
    errorexit ("\njigdo-port: could not open template `%s'.\n\n", template);

  /* skip first 3 lines */
  while (linesread < 3)
    {
      if (fread (&c1, 1, 1, templatefile) != 1)
	errorexit
	  ("\njigdo-port: error reading first lines of template `%s'.\n\n",
	   template);
      if (c2 == 0x0d && c1 == 0x0a)
	linesread++;
      c2 = c1;
    }


  /* First create a completely zero-filled image, to reduce fragmentation */

  /* fill buffer, no memset, keep portable */
  for (i = 0; i < BUFSIZE; i++)
    bufcomp[i] = 0;

  tmpimagefile = fopen (imagetmp, "wb");
  if (!tmpimagefile)
    errorexit ("\njigdo-port: could not create temporary image `%s'.\n\n",
	       imagetmp);

  printf ("\n");

  togo = imagelen;
  while (togo)
    {
      if (written > nextwritten)
	{
	  printf ("Preparing temporary image: %3lu of %3lu MB...\r",
		  (unsigned long) (written / ((unsigned long) 1048576)),
		  (unsigned long) (imagelen / ((unsigned long) 1048576)));
	  fflush (stdout);
	  nextwritten += (unsigned long) 1048576;
	}

      part = BUFSIZE;
      if (togo < part)
	part = togo;
      if (fwrite (bufcomp, 1, part, tmpimagefile) != part)
	errorexit
	  ("\n\njigdo-port: couldn't write temporary image `%s', disk full?\n\n",
	   imagetmp);

      togo -= part;
      written += part;
    }

  printf ("Preparing temporary image: %3lu of %3lu MB, done.\n",
	  (unsigned long) (written / ((unsigned long) 1048576)),
	  (unsigned long) (imagelen / ((unsigned long) 1048576)));


  /* Then fill image with template-provided "literal" data */

  desccurr = deschead;
  /* find first type=2 part */
  while (desccurr != NULL && desccurr->type != 2)
    desccurr = desccurr->next;

  if (desccurr)
    if (fseek (tmpimagefile, desccurr->pos, SEEK_SET))
      errorexit ("\njigdo-port: couldn't seek in temporary image `%s'.\n\n",
		 imagetmp);
  byteswrittentothistype2 = 0;


  printf ("Pasting literal data from template...\r");
  fflush (stdout);

  while (desccurr)
    {

      if (bufuncompoccupied == 0 && bufcompavail == 0)
	/* nothing more to write out and more data needed! */
	{
	  if (bytesstillindatapart == 0)
	    /* new datapart needed, should start at current file location */
	    {
	      bufcomp[0] = 0;
	      fread (bufcomp, 1, 4, templatefile);
	      if (bufcomp[0] != 0x44 || bufcomp[1] != 0x41
		  || bufcomp[2] != 0x54 || bufcomp[3] != 0x41)
		errorexit
		  ("\njigdo-port: probably corrupt template `%s', DATA not found.\n\n",
		   template);

	      bytesstillindatapart = readint6 (templatefile) - 16;
	      bytesstillfromdatapart = readint6 (templatefile);

	      /* prepare unzipping */
	      myz.zalloc = (alloc_func) 0;
	      myz.zfree = (free_func) 0;
	      myz.opaque = (voidpf) 0;
	      myz.next_in = Z_NULL;
	      myz.avail_in = 0;
	      myz.next_out = Z_NULL;
	      myz.avail_out = 0;
	      myz.data_type = Z_BINARY;	/* can't hurt */

	      if (inflateInit (&myz) != Z_OK)
		errorexit ("\njigdo-port: inflateInit error: %s\n\n",
			   myz.msg);
	    }

	  /* read data from template, bufcompavail=0 so start at begin */
	  part = BUFSIZE;
	  if (part > bytesstillindatapart)
	    part = bytesstillindatapart;

	  if (fread (bufcomp, 1, part, templatefile) != part)
	    errorexit ("\njigdo-port: couldn't read from template `%s'.\n\n",
		       template);

	  bytesstillindatapart -= part;
	  bufcompavail = part;
	  bufcompstart = bufcomp;
	}


      if (bufcompavail > 0 && bufuncompoccupied == 0)
	/* no data to write out but still something to decompress */
	{
	  /* inflate */
	  myz.next_in = bufcompstart;
	  myz.avail_in = bufcompavail;
	  myz.next_out = bufuncomp;	/* empty -> start at begin */
	  myz.avail_out = BUFSIZE;

	  err = inflate (&myz, Z_SYNC_FLUSH);
	  if (err != Z_OK && err != Z_STREAM_END)
	    errorexit
	      ("\njigdo-port: error during decompression from template `%s'.\n\n",
	       template);

	  bufcompstart = myz.next_in;
	  bufcompavail = myz.avail_in;

	  bufuncompstart = bufuncomp;
	  bufuncompoccupied = BUFSIZE - myz.avail_out;
	  bytesstillfromdatapart -= bufuncompoccupied;

	  /* close if finished with this part */
	  if (err == Z_STREAM_END)
	    {
	      if (bytesstillindatapart > 0)
		errorexit
		  ("\njigdo-port: too much data in DATA part of template `%s'.\n\n",
		   template);
	      if (bytesstillfromdatapart > 0)
		errorexit
		  ("\njigdo-port: not enough data from DATA part of template `%s'.\n\n",
		   template);

	      inflateEnd (&myz);
	    }
	}


      /* write uncompressed data to file */
      if (bufuncompoccupied > 0)
	{
	  part = bufuncompoccupied;
	  if (part > desccurr->len - byteswrittentothistype2)
	    part = desccurr->len - byteswrittentothistype2;

	  if (fwrite (bufuncompstart, 1, part, tmpimagefile) != part)
	    errorexit
	      ("\njigdo-port: couldn't write in temporary image `%s'.\n\n",
	       imagetmp);

	  byteswrittentothistype2 += part;
	  bufuncompstart += part;
	  bufuncompoccupied -= part;
	}


      if (byteswrittentothistype2 >= desccurr->len)
	{
	  /* find next type=2 part */
	  desccurr = desccurr->next;
	  while (desccurr != NULL && desccurr->type != 2)
	    desccurr = desccurr->next;

	  if (desccurr)
	    if (fseek (tmpimagefile, desccurr->pos, SEEK_SET))
	      errorexit
		("\njigdo-port: couldn't seek in temporary image `%s'.\n\n",
		 imagetmp);

	  byteswrittentothistype2 = 0;
	}
    }

  printf ("Pasting literal data from template, done.\n\n");
  fflush (stdout);

  writedesc (tmpimagefile);

  tmpimageavail = 1;
  tmpimageopen = 1;

#undef BUFSIZE
}


void
opentmpimage ()
{
  if (tmpimageopen)
    return;

  if (!tmpimageavail)
    {
      mkfirsttmpimage ();
      return;
    }

  tmpimagefile = fopen (imagetmp, "r+b");
  if (!tmpimagefile)
    errorexit ("\njigdo-port: could not open temporary image `%s'.\n\n",
	       imagetmp);
  tmpimageopen = 1;
}


int
checknpastefile (filename)
     char *filename;
/* returns 0: failed/not needed, >=1: success&pasted N times */
{
#define BUFSIZE 30000
  unsigned char buf[BUFSIZE];
  ULINT filesize = 0;
  ULINT filesizeagain;
  ULINT bytesread = 0;
  struct MD5Context md5ctx;
  struct MD5Context *md5ctxp = &md5ctx;
  FILE *f;
  md5byte thismd5[16];
  md5byte againmd5[16];
  descent *desccurr = NULL;
  int i;
  int res = 0;
  int md5ok;


  printf ("Checking file `%s'...\n", filename);
  fflush (stdout);

  /* Open file and check MD5 */
  if ((f = fopen (filename, "rb")) == NULL)
    {
      printf ("  Unable to open.\n");
      fflush (stdout);
      return 0;
    }

  MD5Init (md5ctxp);

  while ((bytesread = fread (buf, 1, BUFSIZE, f)) > 0)
    {
      MD5Update (md5ctxp, buf, bytesread);
      filesize += bytesread;
    }

  MD5Final (thismd5, md5ctxp);

#if 0
  printf ("\nMD5=");
  for (i = 0; i < 16; i++)
    printf ("%02x", thismd5[i]);
  printf ("\n");
#endif

  /* See if we need this file */
  desccurr = deschead;
  while (desccurr)
    {
      if (desccurr->type == 3 && desccurr->len == filesize)
	{
	  md5ok = 1;
	  for (i = 0; i < 16; i++)
	    if ((desccurr->md5)[i] != thismd5[i])
	      md5ok = 0;
	  if (md5ok)
	    {
	      /* All checked ok, then paste now. */
	      if (!tmpimageopen)
		opentmpimage ();

	      printf (res ==
		      0 ? "  Pasting...\n" : "  Pasting (location %d)...\n",
		      res + 1);
	      fflush (stdout);

	      rewind (f);
	      MD5Init (md5ctxp);
	      filesizeagain = 0;

	      if (fseek (tmpimagefile, desccurr->pos, SEEK_SET))
		errorexit
		  ("\njigdo-port: couldn't seek in temporary image `%s'.\n\n",
		   imagetmp);

	      bytesread = 0;
	      while ((bytesread = fread (buf, 1, BUFSIZE, f)) > 0)
		{
		  if (fwrite (buf, 1, bytesread, tmpimagefile) != bytesread)
		    errorexit
		      ("\njigdo-port: couldn't write in temporary image `%s'.\n\n",
		       imagetmp);

		  MD5Update (md5ctxp, buf, bytesread);
		  filesizeagain += bytesread;
		}

	      if (filesizeagain != filesize)
		errorexit
		  ("\njigdo-port: filesize doesn't match during second read of `%s'.\n\n",
		   filename);

	      MD5Final (againmd5, md5ctxp);
	      md5ok = 1;
	      for (i = 0; i < 16; i++)
		if (againmd5[i] != thismd5[i])
		  md5ok = 0;
	      if (!md5ok)
		errorexit
		  ("\njigdo-port: MD5sum doesn't match during second read of `%s'.\n\n",
		   filename);

	      res++;
	      desccurr->type = 4;
	      writedesc (tmpimagefile);	/* Update status */

	      /* Keep checking, we may need it another time. */
	    }
	}
      desccurr = desccurr->next;
    }

  fclose (f);

  if (res)
    printf ("  Done.\n");
  else
    printf ("  Not needed.\n");
  fflush (stdout);

  return res;

#undef BUFSIZE
}


void
almostbase64tomd5 (base64, md5)
     char *base64;
     md5byte *md5;
{
/* Details in http://www.ietf.org/rfc/rfc2045.txt
   BUT here 62 = '-' and 63 = '_'
*/

  int firstbit, firstbyte;
  int i;
  unsigned char dec = 0;

  /* clean it up first */
  for (i = 0; i < 16; i++)
    md5[i] = 0;

  for (i = 0; i < 22; i++)
    {
      if (base64[i] >= 'A' && base64[i] <= 'Z')
	dec = base64[i] - 'A';
      else if (base64[i] >= 'a' && base64[i] <= 'z')
	dec = base64[i] - 'a' + 26;
      else if (base64[i] >= '0' && base64[i] <= '9')
	dec = base64[i] - '0' + 52;
      else if (base64[i] == '-')
	dec = 62;
      else if (base64[i] == '_')
	dec = 63;
      else
	errorexit ("\njigdo-port: unknown almost-base64 character.\n\n", "");

      firstbyte = (i * 6) / 8;
      firstbit = (i * 6) % 8;

      if (firstbit == 0)
	md5[firstbyte] |= dec << 2;
      else if (firstbit == 6)
	{
	  md5[firstbyte] |= dec >> 4;
	  if (firstbyte < 15)
	    md5[firstbyte + 1] |= dec << 4;
	}
      else if (firstbit == 4)
	{
	  md5[firstbyte] |= dec >> 2;
	  if (firstbyte < 15)
	    md5[firstbyte + 1] |= dec << 6;
	}
      else if (firstbit == 2)
	md5[firstbyte] |= dec;
      else
	errorexit
	  ("\njigdo-port: internal error in almost-base64 decoding.\n\n", "");
    }
}


char *
unquote (c, next)
     char *c;
     char **next;
/* Unquotes, stops at '\0', '\n', '\r' or unquoted '#'/whitespace.
   Returns pointer to unquoted string -- which must be free()d after use!
   If next != NULL then it's set to first unprinted char (if whitespace then
   another argument may follow). */
{
  int quoted = 0;		/* 0 = unquoted, 1 = 'single', 2 = "double" */
  int backslashed = 0;
  char *newtext;
  char *cp = c;
  char *newcp;

/* Function table:  (quo=1 back=1 can't occur)

   char   quo=0 back=0  quo=0 back=1  quo=1 back=0  quo=2 back=0  quo=2 back=1
   ----   ------------  ------------  ------------  ------------  ------------
   '      quo=1         print back=0  quo=0         print         print back=0
   "      quo=2         print back=0  print         quo=0         print back=0
   \      back=1        print back=0  print         back=1        print back=0
   rest   print         print back=0  print         print         print back=0
*/

  /* can't ever be longer than origtext */
  newtext = (char *) malloc (strlen (c) + 1);
  newcp = newtext;

  while (*cp != '\0' && *cp != '\n' && *cp != '\r' &&
	 !((*cp == '#' || *cp == ' ' || *cp == '\t') && !quoted
	   && !backslashed))
    {
      if (*cp == '\'')
	{
	  if (!quoted && !backslashed)
	    quoted = 1;
	  else if (quoted == 1)
	    quoted = 0;
	  else
	    {
	      *newcp = *cp;
	      newcp++;
	      backslashed = 0;
	    }
	}
      else if (*cp == '"')
	{
	  if (!quoted && !backslashed)
	    quoted = 2;
	  else if (quoted == 2 && !backslashed)
	    quoted = 0;
	  else
	    {
	      *newcp = *cp;
	      newcp++;
	      backslashed = 0;
	    }
	}
      else if (*cp == '\\')
	{
	  if (!quoted && !backslashed)
	    backslashed = 1;
	  else if (quoted == 2 && !backslashed)
	    backslashed = 1;
	  else
	    {
	      *newcp = *cp;
	      newcp++;
	      backslashed = 0;
	    }
	}
      else
	{
	  *newcp = *cp;
	  newcp++;
	  backslashed = 0;
	}

      cp++;
    }

  *newcp = '\0';

  if (next != NULL)
    *next = cp;

  return newtext;
}


int
main (argc, argv)
     int argc;
     char *argv[];
{
#define BUFSIZE 30000		/* at least a few hunderd */
  char buf[BUFSIZE];		/* must be system-type char, not forced
				   (un)signed */
  unsigned char bufu[BUFSIZE];
  int argbase;
  int moreopts;
  char *image = NULL;
  char *jigdo = NULL;
  char *filesfrom = NULL;
  int command = CMD_NONE;
  FILE *imagefile = NULL;
  FILE *templatefile = NULL;
  descent *desccurr = NULL;
  descent *descnew;
  ULINT desclen;
  ULINT descpos;
  unsigned char desctype;
  ULINT currpos = 0;
  ULINT part;
  FILE *filesfromfile = NULL;
  int tmpimagefd;
  md5byte checkmd5[16];
  struct MD5Context md5ctx;
  struct MD5Context *md5ctxp = &md5ctx;
  int md5ok;
  int i;
  ULINT readnow;
  ULINT nextread;
  gzFile jigdofile;
  int inparts;
  char *bufptr;
  char *bufptr2;
  uricouple *urinew;
  uricouple *uricurr = NULL;
#ifndef WORDS_BIGENDIAN
  UWORD32 endiancheck = 12;
#endif


#ifndef WORDS_BIGENDIAN
  /* Check correct endianness */
  bufptr = (char *) &endiancheck;
  if (*bufptr != 12)
    errorexit
      ("\njigdo-port: you need to define WORDS_BIGENDIAN in jigdo-port.c and recompile.\n\n",
       "");
#endif

  if (argc < 2)
    errorexit
      ("\njigdo-port: portable subset of jigdo-file.\nSee jigdo-file documentation for usage info.\n\n",
       "");


  /* Parse arguments, inspired by avi2divx */
  argbase = 1;
  moreopts = argc >= 2;

  while (moreopts && argbase < argc)
    {
      if (!strcmp (argv[argbase], "--image") || !strcmp (argv[argbase], "-i"))
	{
	  image = argv[argbase + 1];
	  argbase += 2;
	}
      else if (!strncmp (argv[argbase], "--image=", 8))
	{
	  image = argv[argbase] + 8;
	  argbase += 1;
	}
      else if (!strcmp (argv[argbase], "--jigdo")
	       || !strcmp (argv[argbase], "-j"))
	{
	  jigdo = argv[argbase + 1];
	  argbase += 2;
	}
      else if (!strncmp (argv[argbase], "--jigdo=", 8))
	{
	  jigdo = argv[argbase] + 8;
	  argbase += 1;
	}
      else if (!strcmp (argv[argbase], "--template")
	       || !strcmp (argv[argbase], "-t"))
	{
	  template = argv[argbase + 1];
	  argbase += 2;
	}
      else if (!strncmp (argv[argbase], "--template=", 11))
	{
	  template = argv[argbase] + 11;
	  argbase += 1;
	}
      else if (!strcmp (argv[argbase], "--files-from")
	       || !strcmp (argv[argbase], "-T"))
	{
	  filesfrom = argv[argbase + 1];
	  argbase += 2;
	}
      else if (!strncmp (argv[argbase], "--files-from=", 13))
	{
	  filesfrom = argv[argbase] + 13;
	  argbase += 1;
	}
      else if (!strcmp (argv[argbase], "--uri"))
	/* No short option, and --uri=Label=something not documented */
	{
	  if ((argbase + 1) >= argc)
	    errorexit ("\njigdo-port: --uri needs an argument.\n\n");

	  /* Allocate new struct */
	  urinew = (uricouple *) malloc (sizeof (uricouple));
	  if (!urihead)
	    urihead = urinew;
	  else
	    uricurr->next = urinew;
	  uricurr = urinew;

	  uricurr->next = NULL;
	  uricurr->label = (char *) malloc (strlen (argv[argbase + 1]) + 1);
	  uricurr->value = (char *) malloc (strlen (argv[argbase + 1]) + 1);
	  /* strdup not really portable */
	  strcpy (uricurr->label, argv[argbase + 1]);
	  bufptr = strchr (uricurr->label, '=');
	  if (!bufptr)
	    errorexit ("\njigdo-port: no value for uri label `%s'.\n\n",
		       uricurr->label);
	  *bufptr = '\0';
	  uricurr->value =
	    unquote (strchr (argv[argbase + 1], '=') + 1, NULL);
	  /* remove any "file:" URI */
	  if (!strncmp (uricurr->value, "file:", 5)
	      || !strncmp (uricurr->value, "FILE:", 5))
	    uricurr->value += 5;

	  argbase += 2;
	}
      else if (!strcmp (argv[argbase], "--version")
	       || !strcmp (argv[argbase], "-v"))
	{
	  printf ("jigdo-port version %s\n", VERSION);
	  return 0;
	}
      else if (!strcmp (argv[argbase], "--cache")
	       || !strcmp (argv[argbase], "-c"))
	/* Not implemented, ignorable */
	argbase += 2;
      else if (!strncmp (argv[argbase], "--cache=", 8))
	argbase += 1;
      else if (!strcmp (argv[argbase], "--"))
	{
	  argbase++;
	  moreopts = 0;
	}
      else if (argv[argbase][0] == '-')
	errorexit ("\njigdo-port: option `%s' not implemented.\n\n",
		   argv[argbase]);

      else if (command == CMD_NONE)
	/* doesn't start with '-' so this must be the command if we hadn't
	   seen it already */
	{
	  if (!strcmp (argv[argbase], "make-image")
	      || !strcmp (argv[argbase], "mi"))
	    command = CMD_MAKEIMAGE;
	  else if (!strcmp (argv[argbase], "print-missing")
		   || !strcmp (argv[argbase], "pm"))
	    command = CMD_PRINTMISSING;
	  else if (!strcmp (argv[argbase], "verify")
		   || !strcmp (argv[argbase], "ver"))
	    command = CMD_VERIFY;
	  else
	    errorexit ("\njigdo-port: command `%s' not implemented.\n\n",
		       argv[argbase]);
	  argbase++;
	}
      else
	/* doesn't start with '-' and command already seen, so must be
	   first file name */
	moreopts = 0;
    }

#if 0
  printf ("\njigdo:%s\ntemplate:%s\nimage:%s\nfilesfrom:%s\nargbase:%d\n\n",
	  jigdo, template, image, filesfrom, argbase);
#endif


  if (image && !strcmp (image, "-"))
    errorexit
      ("\njigdo-port: writing image to stdout is not implemented.\n\n", "");


  if (command == CMD_MAKEIMAGE && image)
    if ((imagefile = fopen (image, "rb")) != NULL)
      {
	fclose (imagefile);
	errorexit
	  ("\njigdo-port: image `%s' already exists, no action taken.\n\n",
	   image);
      }


  if (image)
    {
      imagetmp = (char *) malloc (strlen (image) + 10);
      strcpy (imagetmp, image);
      strcat (imagetmp, ".tmp");
    }


  /* All commands need the template DESC data */
  /* try image.tmp first, if available then that's all we need */
  if (imagetmp)
    templatefile = fopen (imagetmp, "rb");
  if (templatefile)
    tmpimageavail = 1;
  if (!templatefile && template)
    templatefile = fopen (template, "rb");
  if (!templatefile)
    errorexit ("\njigdo-port: could not open template.\n\n", "");

  /* length of DESC part */
  if (fseek (templatefile, -6, SEEK_END))
    errorexit ("\njigdo-port: can't seek in %s.\n\n",
	       tmpimageavail ? "temporary image" : "template");
  desclen = readint6 (templatefile);

  if (fseek (templatefile, -desclen, SEEK_END))
    errorexit
      ("\njigdo-port: probably corrupt %s, desclen too long.\n\n",
       tmpimageavail ? "temporary image" : "template");

  buf[0] = 0;
  fread (buf, 1, 4, templatefile);
  if (buf[0] != 0x44 || buf[1] != 0x45 || buf[2] != 0x53 || buf[3] != 0x43)
    errorexit ("\njigdo-port: probably corrupt %s, DESC not found.\n\n",
	       tmpimageavail ? "temporary image" : "template");

  if (desclen != readint6 (templatefile))
    errorexit
      ("\njigdo-port: probably corrupt %s, desclens don't match.\n\n",
       tmpimageavail ? "temporary image" : "template");

  descpos = 16;
  while (descpos < desclen)
    {
      descpos += 23;
      if (!fread (&desctype, 1, 1, templatefile))
	errorexit
	  ("\njigdo-port: probably corrupt %s, couldn't read desctype.\n\n",
	   tmpimageavail ? "temporary image" : "template");

      if (desctype > 1)
	{
	  /* Allocate new struct */
	  descnew = (descent *) malloc (sizeof (descent));
	  if (!deschead)
	    deschead = descnew;
	  else
	    desccurr->next = descnew;
	  desccurr = descnew;

	  desccurr->next = NULL;
	  desccurr->type = desctype;
	  desccurr->len = readint6 (templatefile);
	  desccurr->pos = currpos;
	  currpos += desccurr->len;
	  if (desctype > 2)
	    fread (desccurr->md5, 1, 16, templatefile);
	  else
	    descpos -= 16;
	}
      else if (desctype == 1)
	{
	  imagelen = readint6 (templatefile);
	  fread (imagemd5, 1, 16, templatefile);
	}
    }

  fclose (templatefile);

  if (imagelen != currpos)
    errorexit
      ("\njigdo-port: probably corrupt %s, currpos isn't imagelen.\n\n",
       tmpimageavail ? "temporary image" : "template");

#if 0
  desccurr = deschead;
  while (desccurr)
    {
      printf ("type: %d, pos: %ld, len: %ld\n", (int) desccurr->type,
	      (long) desccurr->pos, (long) desccurr->len);
      desccurr = desccurr->next;
    }
  printf ("imagelen: %ld, currpos: %ld\n", (long) imagelen, (long) currpos);
#endif


/* ========================================================================= */


  if (command == CMD_MAKEIMAGE)
    {
      /* First check all commandline args */
      while (argbase < argc)
	{
	  checknpastefile (argv[argbase]);
	  argbase++;
	}

      /* Then filesfrom */
      if (filesfrom)
	{
	  if (!strcmp (filesfrom, "-"))
	    filesfromfile = stdin;
	  else if ((filesfromfile = fopen (filesfrom, "r")) == NULL)
	    errorexit
	      ("\njigdo-port: couldn't open files-from file `%s'.\n\n",
	       filesfrom);

	  buf[0] = 'a';
	  while (buf[0] != '\0')
	    {
	      buf[0] = '\0';
	      if (fgets (buf, BUFSIZE, filesfromfile))
		{
		  buf[strlen (buf) - 1] = '\0';	/* remove newline */
		  if (buf[0] != '\0')
		    checknpastefile (buf);
		}
	    }
	}

      if (tmpimageopen)
	{
	  /* If still open, then something was written/changed */
	  /* Check if finished */
	  desccurr = deschead;
	  moreopts = 0;		/* or rather "more files needed" */
	  while (desccurr && !moreopts)
	    {
	      if (desccurr->type == 3)
		moreopts = 1;
	      desccurr = desccurr->next;
	    }

	  if (!moreopts)
	    {
#ifdef HAVE_FTRUNCATE
	      fclose (tmpimagefile);
	      tmpimageopen = 0;

#ifndef O_LARGEFILE
#define O_LARGEFILE 0
#endif
	      /* fileno() not really portable, so open seperately */
	      tmpimagefd = open (imagetmp, O_RDWR | O_LARGEFILE);
	      ftruncate (tmpimagefd, imagelen);
	      close (tmpimagefd);

	      rename (imagetmp, image);
#else
	      rename (imagetmp, image);
	      printf
		("\nFinished! However, the image still contains some trailing data that could\n");
	      printf
		("not be removed because you lack the ftruncate() function. This is usually\n");
	      printf
		("no problem (it's still a valid image), but if you are a perfectionist, you\n");
	      printf ("can do the following:\n");
	      printf ("  head -c %ld %s > %s.perfect\n\n", imagelen, image,
		      image);
#endif
	      return 0;
	    }

	  fclose (tmpimagefile);
	  return 1;
	}
      return 1;
    }


/* ========================================================================= */


  if (command == CMD_PRINTMISSING)
    {
      if ((jigdofile = gzopen (jigdo, "rb")) == NULL)
	errorexit ("\njigdo-port: couldn't open .jigdo file `%s'.\n\n",
		   jigdo);

      inparts = 0;
      while (gzgets (jigdofile, buf, BUFSIZE) != Z_NULL)
	{
	  bufptr = buf;
	  while (*bufptr == ' ' || *bufptr == '\t')
	    bufptr++;
	  if (*bufptr == '[')
	    inparts = 0;
	  if (!strncmp (bufptr, "[Parts]", 7))
	    inparts = 1;
	  if (inparts && *bufptr != '\0' && *bufptr != '[' && *bufptr != '#'
	      && *bufptr != '\n' && *bufptr != '\r')
	    {
	      almostbase64tomd5 (bufptr, checkmd5);

#if 0
	      printf ("\nMD5=");
	      for (i = 0; i < 16; i++)
		printf ("%02x", checkmd5[i]);
	      printf ("\n");
#endif
	      /* see if we need this one */
	      desccurr = deschead;
	      moreopts = 0;	/* misuse for "already printed" */
	      while (desccurr)
		{

/* Note: this prints only one possible file, even if more with same md5sum
   are known. If you want to have them all, add "|| desccurr->type == 5" */
		  if (desccurr->type == 3)
		    {
		      md5ok = 1;
		      for (i = 0; i < 16; i++)
			if ((desccurr->md5)[i] != checkmd5[i])
			  md5ok = 0;
		      if (md5ok)
			{
			  if (!moreopts)	/* i.e. !alreadyprinted */
			    {
			      bufptr2 = bufptr + 22;	/* could be '=' */
			      while (*bufptr2 != '=' && *bufptr2 != '\0')
				bufptr2++;
			      /* now on '=' or '\0' */
			      if (*bufptr2 != '\0')
				bufptr2++;
			      /* search first real char */
			      while (*bufptr2 == ' ' || *bufptr2 == '\t')
				bufptr2++;
			      /* now on first real char or '\0' */
			      if (!*bufptr2)
				errorexit
				  ("\njigdo-port: in .jigdo no value for label `%s'.\n\n",
				   bufptr);

			      bufptr = strchr (bufptr2, ':');
			      /* Maybe not in first argument, but that's hard
			         to check. Will give --uri-not-found error. */

			      if (bufptr && strncmp (bufptr2, "http:", 5)
				  && strncmp (bufptr2, "HTTP:", 5)
				  && strncmp (bufptr2, "ftp:", 4)
				  && strncmp (bufptr2, "FTP:", 4)
				  && strncmp (bufptr2, "file:", 5)
				  && strncmp (bufptr2, "FILE:", 5))
				/* Need to lookup uri-couple */
				{
				  *bufptr = '\0';
				  bufptr++;

				  uricurr = urihead;
				  while (uricurr)
				    {
				      if (!strcmp (uricurr->label, bufptr2))
					{
					  printf ("%s", uricurr->value);
					  /* unquoted during arg-parse */
					  break;
					}
				      uricurr = uricurr->next;
				    }

				  if (!uricurr)
				    errorexit
				      ("\njigdo-port: label `%s' not found, specify --uri for it.\n\n",
				       bufptr2);
				}
			      else
				/* No ':' (assume local file) or full URI */
				/* There _may_ also be a ':' in a local file
				   name that does not have a 'file://', which
				   will produce a --uri-not-found error */
				bufptr = bufptr2;

			      bufptr2 = unquote (bufptr, NULL);	/* mallocs */
			      printf ("%s\n", bufptr2);
			      free (bufptr2);

			      moreopts = 1;
			    }
			  desccurr->type = 5;	/* we have it */
			}
		    }

		  desccurr = desccurr->next;
		}
	    }
	}

      gzclose (jigdofile);

      /* see if anything needed that we haven't found yet */
      desccurr = deschead;
      moreopts = 0;		/* or rather "more locations needed" */
      while (desccurr && !moreopts)
	{
	  if (desccurr->type == 3)
	    moreopts = 1;
	  desccurr = desccurr->next;
	}

      if (moreopts)
	errorexit
	  ("\njigdo-port: more locations needed than specified in .jigdo file `%s'.\n\n",
	   jigdo);

      return 0;
    }


/* ========================================================================= */


  if (command == CMD_VERIFY)
    {

      /* misuse tmpimage for _real_ image */
      if ((tmpimagefile = fopen (image, "rb")) == NULL)
	errorexit ("\njigdo-port: couldn't open image `%s'.\n\n", image);

      MD5Init (md5ctxp);

      currpos = imagelen;
      readnow = 0;
      nextread = 0;
      while (currpos > 0)
	{
	  if (readnow >= nextread)
	    {
	      printf ("Checking image: %3lu of %3lu MB...\r",
		      (unsigned long) (readnow / ((unsigned long) 1048576)),
		      (unsigned long) (imagelen / ((unsigned long) 1048576)));
	      fflush (stdout);
	      nextread += (unsigned long) 1048576;
	    }

	  part = BUFSIZE;
	  if (currpos < part)
	    part = currpos;
	  if (fread (bufu, 1, part, tmpimagefile) != part)
	    errorexit ("\njigdo-port: error while reading image `%s'.\n\n",
		       image);

	  MD5Update (md5ctxp, bufu, part);
	  currpos -= part;
	  readnow += part;
	}

      MD5Final (checkmd5, md5ctxp);

      printf ("Checking image: %3lu of %3lu MB, done:\n",
	      (unsigned long) (readnow / ((unsigned long) 1048576)),
	      (unsigned long) (imagelen / ((unsigned long) 1048576)));

#if 0
      printf ("\nMD5=");
      for (i = 0; i < 16; i++)
	printf ("%02x", checkmd5[i]);
      printf ("\n");
#endif

      md5ok = 1;
      for (i = 0; i < 16; i++)
	if (imagemd5[i] != checkmd5[i])
	  md5ok = 0;
      if (md5ok)
	{
	  printf ("OK: Checksum matches, image is good!\n");
	  if (fread (bufu, 1, 1, tmpimagefile) > 0)
	    {
	      printf
		("\nHowever, the image still contains some trailing data that was not removed\n");
	      printf
		("probably because you lack the ftruncate() function. This is usually no\n");
	      printf
		("problem (it's still a valid image), but if you are a perfectionist, you can\n");
	      printf ("do the following:\n");
	      printf ("  head -c %ld %s > %s.perfect\n\n", imagelen, image,
		      image);
	    }
	  fclose (tmpimagefile);
	  return 0;
	}
      else
	errorexit
	  ("ERROR: Checksum does not match, image might be corrupted!\n", "");
    }


/* ========================================================================= */


  errorexit ("\njigdo-port: internal command switch error.\n\n", "");

  return 3;
#undef BUFSIZE
}
